"""
Builds the aircraft code dict
"""

# stdlib
import json
import re
from pathlib import Path

# library
import httpx


URL = "https://en.wikipedia.org/wiki/List_of_ICAO_aircraft_type_designators"
OUTPUT_PATH = Path(__file__).parent / "files" / "aircraft.json"
TAG_PATTERN = re.compile("<[^>]*>")


def main() -> int:
    """Builds/updates aircraft.json codes"""
    resp = httpx.get(URL)
    if resp.status_code != 200:
        return 1
    text = resp.text
    text = text[text.find("<caption>ICAO") :]
    rows = text[: text.find("</table>")].split("<tr>")[2:]
    craft = {}
    for row in rows:
        cols = row.split("\n")
        name = TAG_PATTERN.sub("", cols[3]).strip()
        if "deprecated" in name:
            continue
        code = TAG_PATTERN.sub("", cols[1]).strip()
        if code not in craft:
            craft[code] = name
    json.dump(craft, OUTPUT_PATH.open("w", encoding="utf8"))
    return 0


if __name__ == "__main__":
    main()



"""
Classes for bulk fetching report strings
"""

# pylint: disable=invalid-name

# stdlib
import asyncio as aio
from contextlib import suppress
from typing import List

from avwx.service.base import CallsHTTP, Service


class NOAA_Bulk(Service, CallsHTTP):
    """Subclass for extracting current reports from NOAA CSV files"""

    url = "https://aviationweather.gov/adds/dataserver_current/current/{}s.cache.csv"
    _valid_types = ("metar", "taf", "aircraftreport", "airsigmet")
    _rtype_map = {"airep": "aircraftreport", "pirep": "aircraftreport"}
    _targets = {"aircraftreport": -2}  # else 0

    def __init__(self, report_type: str):
        super().__init__(self._rtype_map.get(report_type, report_type))

    @staticmethod
    def _clean_report(report: str) -> str:
        for remove in (r"\x07", "\x07"):
            report = report.replace(remove, " ")
        return " ".join(report.split())

    def _extract(self, raw: str) -> List[str]:
        reports = []
        index = self._targets.get(self.report_type, 0)
        for line in raw.split("\n")[6:]:
            with suppress(IndexError):
                report = self._clean_report(line.split(",")[index])
                if report:
                    reports.append(report)
        return reports

    def fetch(self, timeout: int = 10) -> List[str]:
        """Bulk fetch report strings from the service"""
        return aio.run(self.async_fetch(timeout))

    async def async_fetch(self, timeout: int = 10) -> List[str]:
        """Asynchronously bulk fetch report strings from the service"""
        url = self.url.format(self.report_type)
        text = await self._call(url, timeout=timeout)
        return self._extract(text)


class NOAA_Intl(Service, CallsHTTP):
    """Bulk scrape international NOAA data"""

    url = "https://www.aviationweather.gov/{}/intl"
    _valid_types = ("airsigmet",)
    _url_map = {"airsigmet": "sigmet"}

    @staticmethod
    def _clean_report(report: str) -> str:
        # for remove in (r"\x07",):
        #     report = report.replace(remove, " ")
        return " ".join(report.split())

    def _extract(self, raw: str) -> List[str]:
        reports = []
        raw = raw[raw.find("<pre>") + 5 :]
        raw = raw[: raw.find("</pre>")]
        for line in raw.split("<br/>"):
            reports.append(self._clean_report(line.strip().strip('"')))
        return reports

    def fetch(self, timeout: int = 10) -> List[str]:
        """Bulk fetch report strings from the service"""
        return aio.run(self.async_fetch(timeout))

    async def async_fetch(self, timeout: int = 10) -> List[str]:
        """Asynchronously bulk fetch report strings from the service"""
        url = self.url.format(self._url_map[self.report_type])
        text = await self._call(url, timeout=timeout)
        return self._extract(text)



"""
Build navaid coordinate map
"""

import json
from pathlib import Path
import httpx

# redirect https://ourairports.com/data/navaids.csv
URL = "https://davidmegginson.github.io/ourairports-data/navaids.csv"
OUTPUT_PATH = Path(__file__).parent / "files" / "navaids.json"


def main():
    """Builds the navaid coordinate map"""
    text = httpx.get(URL).text
    lines = text.strip().split("\n")
    lines.pop(0)
    data = {}
    for line in lines:
        line = line.split(",")
        try:
            ident, lat, lon = line[2].strip('"'), float(line[6]), float(line[7])
        except ValueError:
            continue
        if not ident:
            continue
        try:
            data[ident].add((lat, lon))
        except KeyError:
            data[ident] = {(lat, lon)}
    data = {k: list(v) for k, v in data.items()}
    json.dump(data, OUTPUT_PATH.open("w"), sort_keys=True)


if __name__ == "__main__":
    main()



"""
General aviation report glossary
"""

# Initial reference:
# https://www.redimec.com.ar/contenido/productos/pdf/1421905575_1.pdf

# Duplicate enitries prioritized for METAR
METAR = {"BR": "Mist", "SS": "Sandstorm"}

# Regional locations. Some duplicates
NA_REGIONAL = {
    "AK": "Alaska",
    "AL": "Alabama",
    "ADRNDCK": "Adirondack",
    "ALGHNY": "Allegheny",
    "ALTA": "Alberta",
    "ALUTN": "Aleutian",
    "APLCN": "Appalachian",
    "APLCNS": "Appalachians",
    "AR": "Arkansas",
    "AZ": "Arizona",
    "BAJA": "Baja, California",
    "BC": "British Columbia",
    "BDA": "Bermuda",
    "BLKHLS": "Black Hills",
    "BOOTHEEL": "Bootheel",
    "BRKSHR": "Berkshire",
    "BRKSHRS": "Berkshires",
    "CA": "California",
    "CARIB": "Caribbean",
    "CASCDS": "Cascades",
    "CHSPK": "Chesapeake",
    "CNDN": "Canadian",
    "CO": "Colorado",
    "CONUS": "Continental U.S.",
    "DC": "District of Columbia",
    "DE": "Delaware",
    "DELMARVA": "Delaware-Maryland-Virginia",
    "DKTS": "Dakotas",
    "FL": "Florida",
    "GA": "Georgia",
    "GRTLKS": "Great Lakes",
    "HDSVLY": "Hudson Valley",
    "IA": "Iowa",
    "ID": "Idaho",
    "IL": "Illinois",
    "KS": "Kansas",
    "KY": "Kentucky",
    "LA": "Louisiana",
    "LABRDR": "Labrador",
    "MA": "Massachusetts",
    "MAN": "Manitoba",
    "MD": "Maryland",
    "ME": "Maine",
    "MEX": "Mexico",
    "MHKVLY": "Mohawk Valley",
    "MI": "Michigan",
    "MN": "Minnesota",
    "MO": "Missouri",
    "MS": "Mississippi",
    "MT": "Montana",
    "NB": "New Brunswick",
    "NC": "North Carolina",
    "ND": "North Dakota",
    "NEB": "Nebraska",
    "NEW ENG": "New England",
    "NFLD": "Newfoundland",
    "NH": "New Hampshire",
    "NJ": "New Jersey",
    "NM": "New Mexico",
    "NS": "Nova Scotia",
    "NV": "Nevada",
    "NY": "New York",
    "OH": "Ohio",
    "OK": "Oklahoma",
    "OR": "Oregon",
    "PA": "Pennsylvania",
    "PGTSND": "Puget Sound",
    "PUGET": "Puget Sound",
    "QUE": "Quebec",
    "RCKYS": "Rockies",
    "RI": "Rhode Island",
    "RIOGD": "Rio Grande",
    "SASK": "Saskatchewan",
    "SC": "South Carolina",
    "SD": "South Dakota",
    "SIERNEV": "Sierra Nevada",
    "SPLNS": "Southern Plains",
    "TN": "Tennessee",
    "TX": "Texas",
    "UT": "Utah",
    "VA": "Virginia",
    "VRISL": "Vancouver Island, BC",
    "VT": "Vermont",
    "WA": "Washington",
    "WI": "Wisconsin",
    "WPLTO": "Western Plateau",
    "WSTCH": "Wasatch Range",
    "WV": "West Virginia",
    "WY": "Wyoming",
    "YKN": "Yukon",
    "YLSTN": "Yellowstone",
}

GLOBAL = {
    "ABNDT": "Abundant",
    "ABNML": "Abnormal",
    "ABT": "About",
    "ABV": "Above",
    "AC": "Convective outlook or altocumulus",
    "ACC": "Altocumulus castellanus clouds",
    "ACCAS": "Altocumulus castellanus clouds",
    "ACCUM": "Accumulate",
    "ACFT": "Aircraft",
    "ACFT MSHP": "Aircraft Mishap",
    "ACLT": "Accelerate",
    "ACLTD": "Accelerated",
    "ACLTG": "Accelerating",
    "ACLTS": "Accelerates",
    "ACPY": "Accompany",
    "ACRS": "Across",
    "ACSL": "Altocumulus standing lenticular",
    "ACTV": "Active",
    "ACTVTY": "Activity",
    "ACYC": "Anticyclone",
    "ADJ": "Adjacent",
    "ADL": "Additional",
    "ADQT": "Adequate",
    "ADQTLY": "Adequately",
    "ADVCT": "Advect",
    "ADVCTD": "Advected",
    "ADVCTG": "Advecting",
    "ADVCTN": "Advection",
    "ADVCTS": "Advects",
    "ADVN": "Advance",
    "ADVNG": "Advancing",
    "ADVY": "Advisory",
    "ADVYS": "Advisories",
    "AFCT": "Affect",
    "AFCTD": "Affected",
    "AFCTG": "Affecting",
    "AFDK": "After dark",
    "AFOS": "Automated Field Operations System",
    "AFSS": "Automated Flight Service Station",
    "AFT": "After",
    "AFTN": "Afternoon",
    "AGL": "Above ground level",
    "AGN": "Again",
    "AGRD Agreed": "Agreed",
    "AGRMT Agreement": "Agreement",
    "AGRS Agrees": "Agrees",
    "AHD Ahead": "Ahead",
    "AIRMET": "Airman's Meteorological Information",
    "ALF": "Aloft",
    "ALG": "Along",
    "ALP": "Airport Location Point",
    "ALQDS": "All quadrants",
    "ALSTG": "Altimeter setting",
    "ALT": "Altitude",
    "ALTHO": "Although",
    "ALTM": "Altimeter",
    "AMD": "Amend",
    "AMDD": "Amended",
    "AMDG": "Amending",
    "AMDT": "Amendment",
    "AMP": "Amplify",
    "AMPG": "Amplifying",
    "AMPLTD": "Amplitude",
    "AMS": "Air mass",
    "AMT": "Amount",
    "ANLYS": "Analysis",
    "ANS": "Answer",
    "AO1": "Automated Reporting Station",
    "AO2": "Automated Reporting Station",
    "AOA": "At or above",
    "AOB": "At or below",
    "AP": "Anomalous Propagation",
    "APCH": "Approach",
    "APCHG": "Approaching",
    "APCHS": "Approaches",
    "APPR": "Appear",
    "APPRG": "Appearing",
    "APPRS": "Appears",
    "APRNT": "Apparent",
    "APRNTLY": "Apparently",
    "APRX": "Approximate",
    "APRXLY": "Approximately",
    "ARL": "Air Resources Lab",
    "ARND": "Around",
    "ARPT": "Airport",
    "ASAP": "As soon as possible",
    "ASL": "Above Sea Level",
    "ASMD": "As Amended",
    "ASOS": "Automated Surface Observing System",
    "ASSOCD": "Associated",
    "ASSOCN": "Association",
    "ATCT": "Air Traffic Control Tower",
    "ATLC": "Atlantic",
    "ATTM": "At this time",
    "ATTN": "Attention",
    "AUTO": "Automated report",
    "AVBL": "Available",
    "AVG": "Average",
    "AVN": "Aviation model",
    "AWC": "Aviation Weather Center",
    "AWIPS": "Advanced Interactive Weather Processing System",
    "AWOS": "Automated Weather Observing system",
    "AWT": "Awaiting",
    "AWW": "Alert Weather Watch",
    "AZM": "Azimuth",
    "B": "Began",
    "BACLIN": "Baroclinic",
    "BATROP": "Barotropic",
    "BC": "Patches",
    "BCFG": "Patchy fog",
    "BCH": "Beach",
    "BCKG": "Backing",
    "BCM": "Become",
    "BCMG": "Becoming",
    "BCMS": "Becomes",
    "BD": "Blowing dust",
    "BDRY": "Boundary",
    "BECMG": "Becoming",
    "BFDK": "Before dark",
    "BFR": "Before",
    "BGN": "Begin",
    "BGNG": "Beginning",
    "BGNS": "Begins",
    "BHND": "Behind",
    "BINOVC": "Breaks in overcast",
    "BKN": "Broken",
    "BL": "Blowing",
    "BLD": "Build",
    "BLDG": "Building",
    "BLDS": "Builds",
    "BLDUP": "Buildup",
    "BLKT": "Blanket",
    "BLKTG": "Blanketing",
    "BLKTS": "Blankets",
    "BLO": "Below or below clouds",
    "BLW": "Below",
    "BLZD": "Blizzard",
    "BN": "Blowing sand",
    "BND": "Bound",
    "BNDRY": "Boundary",
    "BNDRYS": "Boundaries",
    "BNTH": "Beneath",
    "BR": "Branch",
    "BRF": "Brief",
    "BRG": "Branching",
    "BRK": "Break",
    "BRKG": "Breaking",
    "BRKHIC": "Breaks in higher clouds",
    "BRKS": "Breaks",
    "BRM": "Barometer",
    "BRN": "Bulk Richardson Number",
    "BRS": "Branches",
    "BS": "Blowing snow",
    "BTWN": "Between",
    "BWER": "Bounded weak echo region",
    "BYD": "Beyond",
    "C": "Celsius",
    "CA": "Cloud-to-air lightning",
    "CAA": "Cold air advection",
    "CAPE": "Convective available potential energy",
    "CAS": "Committee for Aviation Services",
    "CAT": "Clear air turbulence",
    "CAVOK": "Ceiling and visibility OK",
    "CAVU": "Ceiling and visibility unlimited",
    "CB": "Cumulonimbus",
    "CBMAM": "Cumulonimbus Mammatus clouds",
    "CC": "Cirrocumulus",
    "CCL": "Convective condensation level",
    "CCLDS": "Clear of clouds",
    "CCLKWS": "Counterclockwise",
    "CCSL": "Cirrocumulus standing lenticular",
    "CDFNT": "Cold front",
    "CDFNTL": "Cold frontal",
    "CFP": "Cold front passage",
    "CG": "Cloud-to-ground lightning",
    "CHC": "Chance",
    "CHCS": "Chances",
    "CHG": "Change",
    "CHGD": "Changed",
    "CHGG": "Changing",
    "CHGS": "Changes",
    "CHI": "Cloud-Height indicator",
    "CHINO": "Sky condition at secondary location not available",
    "CHOP": "Choppy turbulence",
    "CI": "Cirrus",
    "CIG": "Ceiling",
    "CIGS": "Ceilings",
    "CIN": "Convective inhibition",
    "CLD": "Cloud",
    "CLDNS": "Cloudiness",
    "CLDS": "Clouds",
    "CLKWS": "Clockwise",
    "CLR": "Clear",
    "CLRG": "Clearing",
    "CLRS": "Clears",
    "CMPLX": "Complex",
    "CNCL": "Cancel",
    "CNCLD": "Canceled",
    "CNCLG": "Canceling",
    "CNCLS": "Cancels",
    "CNTR": "Center",
    "CNTRD": "Centered",
    "CNTRL": "Central",
    "CNTRLN": "Centerline",
    "CNTRS": "Centers",
    "CNTY": "County",
    "CNTYS": "Counties",
    "CNVG": "Converge",
    "CNVGG": "Converging",
    "CNVGNC": "Convergence",
    "CNVTN": "Convection",
    "CNVTV": "Convective",
    "CNVTVLY": "Convectively",
    "COMPAR": "Compare",
    "COMPARD": "Compared",
    "COMPARG": "Comparing",
    "COMPARS": "Compares",
    "COMPR": "Compare",
    "COMPRD": "Compared",
    "COMPRG": "Comparing",
    "COMPRS": "Compares",
    "COND": "Condition",
    "CONFDC": "Confidence",
    "CONS": "Continuous",
    "CONT": "Continue",
    "CONTD": "Continued",
    "CONTDVD": "Continental Divide",
    "CONTG": "Continuing",
    "CONTLY": "Continually",
    "CONTRAILS": "Condensation trails",
    "CONTS": "Continues",
    "COORD": "Coordinate",
    "COR": "Correction",
    "CPBL": "Capable",
    "CPC": "Climate Prediction Center",
    "CRC": "Circle",
    "CRCLC": "Circulate",
    "CRCLN": "Circulation",
    "CRLC": "Circulate",
    "CRLN": "Circulation",
    "CRNR": "Corner",
    "CRNRS": "Corners",
    "CRS": "Course",
    "CS": "Cirrostratus",
    "CSDR": "Consider",
    "CSDRBL": "Considerable",
    "CST": "Coast",
    "CSTL": "Coastal",
    "CT": "Connecticut",
    "CTA": "Control area",
    "CTC": "Contact",
    "CTGY": "Category",
    "CTSKLS": "Catskills",
    "CU": "Cumulus",
    "CUFRA": "Cumulus fractus",
    "CVR": "Cover",
    "CVRD": "Covered",
    "CVRG": "Covering",
    "CVRS": "Covers",
    "CWSU": "Center Weather Service Units",
    "CYC": "Cyclonic",
    "CYCLGN": "Cyclogenesis",
    "DABRK": "Daybreak",
    "DALGT": "Daylight",
    "DBL": "Double",
    "DCR": "Decrease",
    "DCRD": "Decreased",
    "DCRG": "Decreasing",
    "DCRGLY": "Decreasingly",
    "DCRS": "Decreases",
    "DEG": "Degree",
    "DEGS": "Degrees",
    "DFCLT": "Difficult",
    "DFCLTY": "Difficulty",
    "DFNT": "Definite",
    "DFNTLY": "Definitely",
    "DFRS": "Differs",
    "DFUS": "Diffuse",
    "DGNL": "Diagonal",
    "DGNLLY": "Diagonally",
    "DIGG": "Digging",
    "DIR": "Direction",
    "DISC": "Discontinue",
    "DISCD": "Discontinued",
    "DISCG": "Discontinuing",
    "DISRE": "Disregard",
    "DISRED": "Disregarded",
    "DISREG": "Disregarding",
    "DLA": "Delay",
    "DLAD": "Delayed",
    "DLT": "Delete",
    "DLTD": "Deleted",
    "DLTG": "Deleting",
    "DLY": "Daily",
    "DMG": "Damage",
    "DMGD": "Damaged",
    "DMGG": "Damaging",
    "DMNT": "Dominant",
    "DMSH": "Diminish",
    "DMSHD": "Diminished",
    "DMSHG": "Diminishing",
    "DMSHS": "Diminishes",
    "DNDFTS": "Downdrafts",
    "DNS": "Dense",
    "DNSLP": "Downslope",
    "DNSTRM": "Downstream",
    "DNWND": "Downwind",
    "DP": "Deep",
    "DPND": "Deepened",
    "DPNG": "Deepening",
    "DPNS": "Deepens",
    "DPR": "Deeper",
    "DPTH": "Depth",
    "DR": "Low Drifting",
    "DRDU": "Drifting dust",
    "DRFT": "Drift",
    "DRFTD": "Drifted",
    "DRFTG": "Drifting",
    "DRFTS": "Drifts",
    "DRSA": "Low drifting sand",
    "DRSN": "Low drifting snow",
    "DRZL": "Drizzle",
    "DS": "Duststorm",
    "DSCNT": "Descent",
    "DSIPT": "Dissipate",
    "DSIPTD": "Dissipated",
    "DSIPTG": "Dissipating",
    "DSIPTN": "Dissipation",
    "DSIPTS": "Dissipates",
    "DSND": "Descend",
    "DSNDG": "Descending",
    "DSNDS": "Descends",
    "DSNT": "Distant",
    "DSTBLZ": "Destabilize",
    "DSTBLZD": "Destabilized",
    "DSTBLZG": "Destabilizing",
    "DSTBLZN": "Destabilization",
    "DSTBLZS": "Destabilizes",
    "DSTC": "Distance",
    "DTRT": "Deteriorate",
    "DTRTD": "Deteriorated",
    "DTRTG": "Deteriorating",
    "DTRTS": "Deteriorates",
    "DU": "Widespread dust storm",
    "DURC": "During climb",
    "DURD": "During descent",
    "DURG": "During",
    "DURGC": "During climb",
    "DURGD": "During descent",
    "DURN": "Duration",
    "DVLP": "Develop",
    "DVLPD": "Developed",
    "DVLPG": "Developing",
    "DVLPMT": "Development",
    "DVLPS": "Develops",
    "DVRG": "Diverge",
    "DVRGG": "Diverging",
    "DVRGNC": "Divergence",
    "DVRGS": "Diverges",
    "DVV": "Downward vertical velocity",
    "DWNDFTS": "Downdrafts",
    "DWPNT": "Dew point",
    "DWPNTS": "Dew points",
    "DX": "Duplex",
    "DZ": "Drizzle",
    "E": "East",
    "EBND": "Eastbound",
    "EFCT": "Effect",
    "ELNGT": "Elongate",
    "ELNGTD": "Elongated",
    "ELSW": "Elsewhere",
    "EMBD": "Embedded",
    "EMBDD": "Embedded",
    "EMERG": "Emergency",
    "ENCTR": "Encounter",
    "ENDG": "Ending",
    "ENE": "East-northeast",
    "ENELY": "East-northeasterly",
    "ENERN": "East-northeastern",
    "ENEWD": "East-northeastward",
    "ENHNC": "Enhance",
    "ENHNCD": "Enhanced",
    "ENHNCG": "Enhancing",
    "ENHNCMNT": "Enhancement",
    "ENHNCS": "Enhances",
    "ENRT": "Enroute",
    "ENTR": "Entire",
    "ERN": "Eastern",
    "ERY": "Early",
    "ERYR": "Earlier",
    "ESE": "East-southeast",
    "ESELY": "East-southeasterly",
    "ESERN": "East-southeastern",
    "ESEWD": "East-southeastward",
    "ESNTL": "Essential",
    "EST": "Estimate",
    "ESTAB": "Establish",
    "ESTS": "Estimates",
    "ETA": "Estimated time of arrival",
    "ETC": "Et cetera",
    "ETIM": "Elapsed time",
    "EVE": "Evening",
    "EWD": "Eastward",
    "EXCLV": "Exclusive",
    "EXCLVLY": "Exclusively",
    "EXCP": "Except",
    "EXP": "Expected",
    "EXPC": "Expect",
    "EXPCD": "Expected",
    "EXPCG": "Expecting",
    "EXTD": "Extend",
    "EXTDD": "Extended",
    "EXTDG": "Extending",
    "EXTDS": "Extends",
    "EXTN": "Extension",
    "EXTRAP": "Extrapolate",
    "EXTRAPD": "Extrapolated",
    "EXTRM": "Extreme",
    "EXTRMLY": "Extremely",
    "EXTSV": "Extensive",
    "F": "Fahrenheit",
    "FA": "Aviation area forecast",
    "FAH": "Fahrenheit",
    "FAM": "Familiar",
    "FC": "Funnel cloud",
    "FCST": "Forecast",
    "FCSTD": "Forecasted",
    "FCSTG": "Forecasting",
    "FCSTR": "Forecaster",
    "FCSTS": "Forecasts",
    "FEW": "Few",
    "FG": "Fog",
    "FIBI": "Filed but impracticable to transmit",
    "FIG": "Figure",
    "FILG": "Filling",
    "FIR": "Flight information region",
    "FIRAV": "First available",
    "FIRST": "First observation after a break in coverage at manual station",
    "FIS": "Flight Information Service",
    "FIS-B": "Flight Information Service Broadcast",
    "FL": "Flight level",
    "FLG": "Falling",
    "FLRY": "Flurry",
    "FLRYS": "Flurries",
    "FLT": "Flight",
    "FLW": "Follow",
    "FLWG": "Following",
    "FM": "From",
    "FMT": "Format",
    "FNCTN": "Function",
    "FNT": "Front",
    "FNTGNS": "Frontogenesis",
    "FNTL": "Frontal",
    "FNTLYS": "Frontolysis",
    "FNTS": "Fronts",
    "FORNN": "Forenoon",
    "FPM": "Feet per minute",
    "FQT": "Frequent",
    "FQTLY": "Frequently",
    "FRM": "Form",
    "FRMG": "Forming",
    "FRMN": "Formation",
    "FROPA": "Frontal passage",
    "FROSFC": "Frontal surface",
    "FRQ": "Frequent",
    "FRST": "Frost",
    "FRWF": "Forecast wind factor",
    "FRZ": "Freeze",
    "FRZG": "Freezing",
    "FRZLVL": "Freezing level",
    "FRZN": "Frozen",
    "FT": "Feet or Terminal Forecast",
    "FTHR": "Further",
    "FU": "Smoke",
    "FV": "Flight visibility",
    "FVRBL": "Favorable",
    "FWD": "Forward",
    "FYI": "For your information",
    "FZ": "Freezing",
    "FZRANO": "Freezing rain sensor not available",
    "G": "Gust",
    "GEN": "General",
    "GENLY": "Generally",
    "GEO": "Geographic",
    "GEOREF": "Geographical reference",
    "GF": "Fog",
    "GICG": "Glaze icing",
    "GLFALSK": "Gulf of Alaska",
    "GLFCAL": "Gulf of California",
    "GLFMEX": "Gulf of Mexico",
    "GLFSTLAWR": "Gulf of St. Lawrence",
    "GND": "Ground",
    "GNDFG": "Ground fog",
    "GOES": "Geostationary Operational Environmental Satellite",
    "GR": "Hail",
    "GRAD": "Gradient",
    "GRDL": "Gradual",
    "GRDLY": "Gradually",
    "GRT": "Great",
    "GRTLY": "Greatly",
    "GRTR": "Greater",
    "GRTST": "Greatest",
    "GS": "Small hail",
    "GSTS": "Gusts",
    "GSTY": "Gusty",
    "GTS": "Global Telecommunication System",
    "GV": "Ground visibility",
    "HAZ": "Hazard",
    "HCVIS": "High clouds visible",
    "HDFRZ": "Hard freeze",
    "HDWND": "Head wind",
    "HGT": "Height",
    "HI": "High or Hawaii",
    "HIER": "Higher",
    "HIFOR": "High level forecast",
    "HLF": "Half",
    "HLSTO": "Hailstones",
    "HLTP": "Hilltop",
    "HLYR": "Haze layer",
    "HND": "Hundred",
    "HPC": "Hydrometeorological Prediction Center",
    "HR": "Hour",
    "HRS": "Hours",
    "HRZN": "Horizon",
    "HTG": "Heating",
    "HURCN": "Hurricane",
    "HUREP": "Hurricane report",
    "HV": "Have",
    "HVY": "Heavy",
    "HVYR": "Heavier",
    "HVYST": "Heaviest",
    "HWVR": "However",
    "HWY": "Highway",
    "HZ": "Haze",
    "IC": "Ice crystals or ice",
    "ICAO": "International Civil Aviation Organization",
    "ICG": "Icing",
    "ICGIC": "Icing in clouds",
    "ICGICIP": "Icing in clouds and in precipitation",
    "ICGIP": "Icing in precipitation",
    "IFR": "Instrument flight rules",
    "IMC": "Instrument meteorolgical conditions",
    "IMDT": "Immediate",
    "IMDTLY": "Immediately",
    "IMPL": "Impulse",
    "IMPLS": "Impulses",
    "IMPT": "Important",
    "INCL": "Include",
    "INCLD": "Included",
    "INCLG": "Including",
    "INCLS": "Includes",
    "INCR": "Increase",
    "INCRD": "Increased",
    "INCRG": "Increasing",
    "INCRGLY": "Increasingly",
    "INCRS": "Increases",
    "INDC": "Indicate",
    "INDCD": "Indicated",
    "INDCG": "Indicating",
    "INDCS": "Indicates",
    "INDEF": "Indefinite",
    "INFO": "Information",
    "INLD": "Inland",
    "INSTBY": "Instability",
    "INTCNTL": "Intercontinental",
    "INTER": "Intermittent",
    "INTL": "International",
    "INTMD": "Intermediate",
    "INTMT": "Intermittent",
    "INTMTLY": "Intermittently",
    "INTR": "Interior",
    "INTRMTRGN": "Intermountain region",
    "INTS": "Intense",
    "INTSFCN": "Intensification",
    "INTSFY": "Intensify",
    "INTSFYD": "Intensified",
    "INTSFYG": "Intensifying",
    "INTSFYS": "Intensifies",
    "INTSTY": "Intensity",
    "INTVL": "Interval",
    "INVOF": "In vicinity of",
    "INVRN": "Inversion",
    "IOVC": "In overcast",
    "IP": "Ice pellets",
    "IPV": "Improve",
    "IPVG": "Improving",
    "IR": "Infrared",
    "ISOL": "Isolate",
    "ISOLD": "Isolated",
    "JCTN": "Junction",
    "JTSTR": "Jet stream",
    "KFRST": "Killing frost",
    "KLYR": "Smoke layer aloft",
    "KMH": "Kilometers per hour",
    "KOCTY": "Smoke over city",
    "KT": "Knots",
    "L": "Left",
    "LAMP": "Local AWIPS MOS Program",
    "LAPS": "Local Analysis and Prediction System",
    "LAST": "Last observation before a break in coverage at a manual station",
    "LAT": "Latitude",
    "LAWRS": "Limited aviation weather reporting station",
    "LCL": "Local or Lifted condensation level",
    "LCLY": "Locally",
    "LCTD": "Located",
    "LCTMP": "Little change in temperature",
    "LCTN": "Location",
    "LDG": "Landing",
    "LEVEL": "Level",
    "LFM": "Limited fine mesh model",
    "LFTG": "Lifting",
    "LGRNG": "Long-range",
    "LGT": "Light",
    "LGTR": "Lighter",
    "LGWV": "Long wave",
    "LI": "Lifted Index",
    "LIFR": "Low instrument flight rules",
    "LIS": "Lifted Indices",
    "LK": "Lake",
    "LKLY": "Likely",
    "LKS": "Lakes",
    "LLJ": "Low level jet",
    "LLWAS": "Low-level wind shear alert system",
    "LLWS": "Low-level wind shear",
    "LMTD": "Limited",
    "LMTG": "Limiting",
    "LMTS": "Limits",
    "LN": "Line",
    "LNS": "LInes",
    "LO": "Low",
    "LONG": "Longitude",
    "LONGL": "Longitudinal",
    "LRG": "Large",
    "LRGLY": "Largely",
    "LRGR": "Larger",
    "LRGST": "Largest",
    "LST": "Local standard time",
    "LTD": "Limited",
    "LTG": "Lightning",
    "LTGCA": "Lightning cloud-to-air",
    "LTGCC": "Lightning cloud-to-cloud",
    "LTGCCCG": "Lightning cloud-to-cloud",
    "LTGCG": "Lightning cloud-to-ground",
    "LTGCW": "Lightning cloud-to-water",
    "LTGIC": "Lightning in cloud",
    "LTL": "Little",
    "LTLCG": "Little change",
    "LTR": "Later",
    "LTST": "Latest",
    "LV": "Leaving",
    "LVL": "Level",
    "LVLS": "Levels",
    "LWR": "Lower",
    "LWRD": "Lowered",
    "LWRG": "Lowering",
    "LYR": "Layer",
    "LYRD": "Layered",
    "LYRS": "Layers",
    "M": "Minus or Less than lowest sensor value",
    "MAX": "Maximum",
    "MB": "Millibars",
    "MCD": "Mesoscale discussion",
    "MDFY": "Modify",
    "MDFYD": "Modified",
    "MDFYG": "Modifying",
    "MDL": "Model",
    "MDLS": "Models",
    "MDT": "Moderate",
    "MDTLY": "Moderately",
    "MED": "Medium",
    "MEGG": "Merging",
    "MESO": "Mesoscale",
    "MET": "Meteorological",
    "METAR": "Aviation Routine Weather Report",
    "METRO": "Metropolitan",
    "MI": "Mile",
    "MID": "Middle",
    "MIDN": "Midnight",
    "MIFG": "Shallow fog",
    "MIL": "Military",
    "MIN": "Minimum",
    "MISG": "Missing",
    "MLTLVL": "Melting level",
    "MNLD": "Mainland",
    "MNLY": "Mainly",
    "MOD": "Moderate",
    "MOGR": "Moderate or greater",
    "MOS": "Model Output Statistics",
    "MOV": "Move",
    "MOVD": "Moved",
    "MOVG": "Moving",
    "MOVMT": "Movement",
    "MOVS": "Moves",
    "MPH": "Miles per hour",
    "MPS": "Meters per second",
    "MRGL": "Marginal",
    "MRGLLY": "Marginally",
    "MRNG": "Morning",
    "MRTM": "Maritime",
    "MSG": "Message",
    "MSL": "Mean sea level",
    "MST": "Most",
    "MSTLY": "Mostly",
    "MSTR": "Moisture",
    "MT": "Mountain",
    "MTN": "Mountain",
    "MTNS": "Mountains",
    "MTW": "Mountain wave",
    "MULT": "Multiple",
    "MULTILVL": "Multilevel",
    "MVFR": "Marginal visual flight rules",
    "MWO": "Meteorological Watch Office",
    "MX": "Mixed snow and ice",
    "MXD": "Mixed",
    "N": "North",
    "N/A": "Not applicable",
    "NAB": "Not above",
    "NAT": "North Atlantic",
    "NATL": "National",
    "NAV": "Navigation",
    "NAVAID": "Electronic navigation aid facility",
    "NBND": "Northbound",
    "NBRHD": "Neighborhood",
    "NC": "No change",
    "NCDC": "National Climatic Data Center",
    "NCEP": "National Center of Environmental Prediction",
    "NCO": "NCEP Central Operations",
    "NCWX": "No change in weather",
    "NE": "Northeast",
    "NEC": "Necessary",
    "NEG": "Negative",
    "NEGLY": "Negatively",
    "NELY": "Northeasterly",
    "NERN": "Northeastern",
    "NEWD": "Northeastward",
    "NGM": "Nested grid model",
    "NGT": "Night",
    "NHC": "National Hurricane Center",
    "NIL": "None",
    "NL": "No layers",
    "NLT": "Not later than",
    "NLY": "Northerly",
    "NM": "Nautical miles",
    "NMBR": "Number",
    "NMBRS": "Numbers",
    "NMC": "National Meteorological Center",
    "NML": "Normal",
    "NMRS": "Numerous",
    "NNE": "North-northeast",
    "NNELY": "North-northeasterly",
    "NNERN": "North-northeastern",
    "NNEWD": "North-northeastward",
    "NNNN": "End of message",
    "NNW": "North-northwest",
    "NNWLY": "North-northwesterly",
    "NNWRN": "North-northwestern",
    "NNWWD": "North-northwestward",
    "NOAA": "National Oceanic and Atmospheric Administration",
    "NOPAC": "Northern Pacific",
    "NOS": "National Ocean Service",
    "NOSPECI": "No SPECI reports are taken at station",
    "NPRS": "Nonpersistent",
    "NR": "Near",
    "NRLY": "Nearly",
    "NRN": "Northern",
    "NRW": "Narrow",
    "NSC": "No significant cloud",
    "NSW": "No significant weather",
    "NTFY": "Notify",
    "NTFYD": "Notified",
    "NVA": "Negative vorticity advection",
    "NW": "Northwest",
    "NWD": "Northward",
    "NWLY": "Northwesterly",
    "NWRN": "Northwestern",
    "NWS": "National Weather Service",
    "NXT": "Next",
    "OAT": "Outside air temperature",
    "OBND": "Outbound",
    "OBS": "Observation",
    "OBSC": "Obscure",
    "OBSCD": "Obscured",
    "OBSCG": "Obscuring",
    "OCFNT": "Occluded front",
    "OCLD": "Occlude",
    "OCLDD": "Occluded",
    "OCLDG": "Occluding",
    "OCLDS": "Occludes",
    "OCLN": "Occlusion",
    "OCNL": "Occasional",
    "OCNLY": "Occasionally",
    "OCR": "Occur",
    "OCRD": "Occurred",
    "OCRG": "Occurring",
    "OCRS": "Occurs",
    "OFC": "Office",
    "OFCM": "Office of the Federal Coordinator for Meteorology",
    "OFP": "Occluded frontal passage",
    "OFSHR": "Offshore",
    "OHD": "Overhead",
    "OMTNS": "Over mountains",
    "ONSHR": "On shore",
    "ORGPHC": "Orographic",
    "ORIG": "Original",
    "OSV": "Ocean station vessel",
    "OTLK": "Outlook",
    "OTP": "On top",
    "OTR": "Other",
    "OTRW": "Otherwise",
    "OUTFLO": "Outflow",
    "OV": "Over",
    "OVC": "Overcast",
    "OVHD": "Overhead",
    "OVNGT": "Overnight",
    "OVR": "Over",
    "OVRN": "Overrun",
    "OVRNG": "Overrunning",
    "OVTK": "Overtake",
    "OVTKG": "Overtaking",
    "OVTKS": "Overtakes",
    "P": "Higher than greatest sensor value",
    "P6SM": "Visibility forecast to be greater than 6 statute miles",
    "PAC": "Pacific",
    "PATWAS": "Pilot's automatic telephone weather answering service",
    "PBL": "Planetary boundary layer",
    "PCPN": "Precipitation",
    "PD": "Period",
    "PDMT": "Predominant",
    "PDS": "Periods",
    "PE": "Ice pellets",
    "PEN": "Peninsula",
    "PERM": "Permanent",
    "PHYS": "Physical",
    "PIBAL": "Pilot balloon observation",
    "PIREP": "Pilot weather report",
    "PK WND": "Peak wind",
    "PL": "Ice pellets",
    "PLNS": "Plains",
    "PLS": "Please",
    "PLTO": "Plateau",
    "PM": "Postmeridian",
    "PNHDL": "Panhandle",
    "PNO": "Precipitation amount not available",
    "PO": "Dust or sand swirls",
    "POS": "Positive",
    "POSLY": "Positively",
    "PPINA": "Radar weather report not available",
    "PPINE": "Radar weather report no echoes observed",
    "PPSN": "Present position",
    "PR": "Partial",
    "PRBL": "Probable",
    "PRBLTY": "Probability",
    "PRBLY": "Probably",
    "PRECD": "Precede",
    "PRECDD": "Preceded",
    "PRECDG": "Preceding",
    "PRECDS": "Precedes",
    "PRES": "Pressure",
    "PRESFR": "Pressure falling rapidly",
    "PRESRR": "Pressure rising rapidly",
    "PRFG": "Partial fog",
    "PRIM": "Primary",
    "PRIN": "Principal",
    "PRIND": "Present indications are",
    "PRJMP": "Pressure jump",
    "PROB": "Probability",
    "PROBC C": "Forecaster's assessment of the probability of occurrence of a thunderstorm or precipitation event, along with associated weather elements (wind, visibility, and/or sky condition) whose occurrences are directly related to, and contemporaneous with, the thunderstorm or precipitation event",
    "PROC": "Procedure",
    "PROD": "Produce",
    "PRODG": "Producing",
    "PROG": "Forecast",
    "PROGD": "Forecasted",
    "PROGS": "Forecasts",
    "PRSNT": "Present",
    "PRSNTLY": "Presently",
    "PRST": "Persist",
    "PRSTNC": "Persistence",
    "PRSTNT": "Persistent",
    "PRSTS": "Persists",
    "PRVD": "Provide",
    "PRVDD": "Provided",
    "PRVDG": "Providing",
    "PRVDS": "Provides",
    "PS": "Plus",
    "PSBL": "Possible",
    "PSBLTY": "Possibility",
    "PSBLY": "Possibly",
    "PSG": "Passage",
    "PSN": "Position",
    "PSND": "Positioned",
    "PTCHY": "Patchy",
    "PTLY": "Partly",
    "PTNL": "Potential",
    "PTNLY": "Potentially",
    "PTNS": "Portions",
    "PVA": "Positive vorticity advection",
    "PVL": "Prevail",
    "PVLD": "Prevailed",
    "PVLG": "Prevailing",
    "PVLS": "Prevails",
    "PVLT": "Prevalent",
    "PWB": "Pilot weather briefing",
    "PWINO": "Precipitation identifier sensor not available",
    "PWR": "Power",
    "PY": "Spray",
    "QN": "Question",
    "QPFERD": "NCEP excessive rainfall discussion",
    "QPFHSD": "NCEP heavy snow discussion",
    "QPFSPD": "NCEP special precipitation discussion",
    "QSTNRY": "Quasistationary",
    "QTR": "Quarter",
    "QUAD": "Quadrant",
    "R": "Right",
    "RA": "Rain",
    "RADAT": "Radiosonde additional data",
    "RAOB": "Radiosonde observation",
    "RCA": "Reach Cruising Altitude",
    "RCH": "Reach",
    "RCHD": "Reached",
    "RCHG": "Reaching",
    "RCHS": "Reaches",
    "RCKY": "Rocky",
    "RCMD": "Recommend",
    "RCMDD": "Recommended",
    "RCMDG": "Recommending",
    "RCMDS": "Recommends",
    "RCRD": "Record",
    "RCRDS": "Records",
    "RCV": "Receive",
    "RCVD": "Received",
    "RCVG": "Receiving",
    "RCVS": "Receives",
    "RDC": "Reduce",
    "RDGG": "Ridging",
    "RDOACT": "Radioactive",
    "RDR": "Radar",
    "RDVLP": "Redevelop",
    "RDVLPG": "Redeveloping",
    "RDVLPMT": "Redevelopment",
    "RE": "Regard",
    "RECON": "Reconnaissance",
    "REF": "Reference",
    "REPL": "Replace",
    "REPLD": "Replaced",
    "REPLG": "Replacing",
    "REPLS": "Replaces",
    "REQ": "Request",
    "REQS": "Requests",
    "REQSTD": "Requested",
    "RES": "Reserve",
    "RESP": "Response",
    "RESTR": "Restrict",
    "RGD": "Ragged",
    "RGL": "Regional model",
    "RGLR": "Regular",
    "RGN": "Region",
    "RGNS": "Regions",
    "RGT": "Right",
    "RH": "Relative humidity",
    "RHINO": "RHI not operative",
    "RIME": "Type of icing characterized by a rough, milky, opaque appearance",
    "RLBL": "Reliable",
    "RLTV": "Relative",
    "RLTVLY": "Relatively",
    "RM": "Remarks",
    "RMK": "Remark",
    "RMN": "Remain",
    "RMND": "Remained",
    "RMNDR": "Remainder",
    "RMNG": "Remaining",
    "RMNS": "Remains",
    "RNFL": "Rainfall",
    "RNG": "Range",
    "ROT": "Rotate",
    "ROTD": "Rotated",
    "ROTG": "Rotating",
    "ROTS": "Rotates",
    "RPD": "Rapid",
    "RPDLY": "Rapidly",
    "RPLC": "Replace",
    "RPLCD": "Replaced",
    "RPLCG": "Replacing",
    "RPLCS": "Replaces",
    "RPRT": "Report",
    "RPRTD": "Reported",
    "RPRTG": "Reporting",
    "RPRTS": "Reports",
    "RPT": "Repeat",
    "RPTG": "Repeating",
    "RPTS": "Repeats",
    "RQR": "Require",
    "RQRD": "Required",
    "RQRG": "Requiring",
    "RQRS": "Requires",
    "RS": "Receiver station",
    "RSG": "Rising",
    "RSN": "Reason",
    "RSNG": "Reasoning",
    "RSNS": "Reasons",
    "RSTR": "Restrict",
    "RSTRD": "Restricted",
    "RSTRG": "Restricting",
    "RSTRS": "Restricts",
    "RTRN": "Return",
    "RTRND": "Returned",
    "RTRNG": "Returning",
    "RTRNS": "Returns",
    "RUC": "Rapid Update Cycle",
    "RUF": "Rough",
    "RUFLY": "Roughly",
    "RVR": "Runway Visual Range",
    "RVRNO": "RVR system not available",
    "RVS": "Revise",
    "RVSD": "Revised",
    "RVSG": "Revising",
    "RVSS": "Revises",
    "RW": "Rain shower",
    "RWY": "Runway",
    "RY": "Runway",
    "S": "South",
    "SA": "Sand",
    "SAB": "Satellite Analysis Branch",
    "SAO": "Surface observation",
    "SATFY": "Satisfactory",
    "SBND": "Southbound",
    "SBSD": "Subside",
    "SBSDD": "Subsided",
    "SBSDNC": "Subsidence",
    "SBSDS": "Subsides",
    "SC": "Stratocumulus",
    "SCND": "Second",
    "SCNDRY": "Secondary",
    "SCSL": "Stratocumulus standing lenticular",
    "SCT": "Scattered",
    "SCTD": "Scattered",
    "SCTR": "Sector",
    "SE": "Southeast",
    "SEC": "Second",
    "SELY": "Southeasterly",
    "SEPN": "Separation",
    "SEQ": "Sequence",
    "SERN": "Southeastern",
    "SEV": "Severe",
    "SEWD": "Southeastward",
    "SFC": "Surface",
    "SFERICS": "Atmospherics",
    "SG": "Snow grains",
    "SGFNT": "Significant",
    "SGFNTLY": "Significantly",
    "SH": "Showers",
    "SHFT": "Shift",
    "SHFTD": "Shifted",
    "SHFTG": "Shifting",
    "SHFTS": "Shifts",
    "SHLD": "Shield",
    "SHLW": "Shallow",
    "SHRT": "Short",
    "SHRTLY": "Shortly",
    "SHRTWV": "Shortwave",
    "SHUD": "Should",
    "SHWR": "Shower",
    "SIG": "Signature",
    "SIGMET": "Significant meteorological information",
    "SIMUL": "Simultaneous",
    "SK": "Sky cover",
    "SKC": "Sky clear",
    "SKED": "Schedule",
    "SLD": "Solid",
    "SLGT": "Slight",
    "SLGTLY": "Slightly",
    "SLO": "Slow",
    "SLOLY": "Slowly",
    "SLOR": "Slower",
    "SLP": "Sea level pressure",
    "SLPG": "Sloping",
    "SLPNO": "Sea-level pressure not available",
    "SLT": "Sleet",
    "SLW": "Slow",
    "SLY": "Southerly",
    "SM": "Statute mile",
    "SMK": "Smoke",
    "SML": "Small",
    "SMLR": "Smaller",
    "SMRY": "Summary",
    "SMS": "Sunchronous meteorological satellite",
    "SMTH": "Smooth",
    "SMTHR": "Smoother",
    "SMTHST": "Smoothest",
    "SMTM": "Sometime",
    "SMWHT": "Somewhat",
    "SN": "Snow",
    "SNBNK": "Snowbank",
    "SND": "Sand",
    "SNFLK": "Snowflake",
    "SNGL": "Single",
    "SNOINCR": "Snow increase",
    "SNOINCRG": "Snow increasing",
    "SNST": "Sunset",
    "SNW": "Snow",
    "SNWFL": "Snowfall",
    "SOP": "Standard operating procedure",
    "SP": "Snow pellets",
    "SPC": "Storm Prediction Center",
    "SPCLY": "Especially",
    "SPD": "Speed",
    "SPECI": "Special observation",
    "SPENES": "Satellite precip. estimate statement",
    "SPKL": "Sprinkle",
    "SPRD": "Spread",
    "SPRDG": "Spreading",
    "SPRDS": "Spreads",
    "SPRL": "Spiral",
    "SQ": "Squall",
    "SQAL": "Squall",
    "SQL": "Squall line",
    "SQLN": "Squall line",
    "SR": "Sunrise",
    "SRN": "Southern",
    "SRND": "Surround",
    "SRNDD": "Surrounded",
    "SRNDG": "Surrounding",
    "SRNDS": "Surrounds",
    "SS": "Sunset",
    "SSE": "South-southeast",
    "SSELY": "South-southeasterly",
    "SSERN": "South-southeastern",
    "SSEWD": "South-southeastward",
    "SSW": "South-southwest",
    "SSWLY": "South-southwesterly",
    "SSWRN": "South-southwestern",
    "SSWWD": "South-southwestward",
    "ST": "Stratus",
    "STAGN": "Stagnation",
    "STBL": "Stable",
    "STBLTY": "Stability",
    "STD": "Standard",
    "STDY": "Steady",
    "STFR": "Stratus fractus",
    "STFRM": "Stratiform",
    "STG": "Strong",
    "STGLY": "Strongly",
    "STGR": "Stronger",
    "STGST": "Strongest",
    "STLT": "Satellite",
    "STM": "Storm",
    "STMS": "Storms",
    "STN": "Station",
    "STNR": "Stationary",
    "STNRY": "Stationary",
    "SUB": "Substitute",
    "SUBTRPCL": "Subtropical",
    "SUF": "Sufficient",
    "SUFLY": "Sufficiently",
    "SUG": "Suggest",
    "SUGG": "Suggesting",
    "SUGS": "Suggests",
    "SUP": "Supply",
    "SUPG": "Supplying",
    "SUPR": "Superior",
    "SUPSD": "Supersede",
    "SUPSDG": "Superseding",
    "SUPSDS": "Supersedes",
    "SVG": "Serving",
    "SVR": "Severe",
    "SVRL": "Several",
    "SW": "Southwest",
    "SW+": "Heavy snow shower",
    "SW-": "Light snow shower",
    "SWD": "Southward",
    "SWLG": "Swelling",
    "SWLY": "Southwesterly",
    "SWODY1": "SPC Severe Weather Outlook for Day 1",
    "SWOMCD": "SPC Mesoscale discussion",
    "SWRN": "Southwestern",
    "SWWD": "Southwestward",
    "SX": "Stability index",
    "SXN": "Section",
    "SYNOP": "Synoptic",
    "SYNS": "Synopsis",
    "SYS": "System",
    "T": "Thunder",
    "TA": "Temperature",
    "TACAN": "UHF Tactical Air Navigation Aid",
    "TAF": "Terminal Area Forecast",
    "TB": "Turbulence",
    "TC": "Tropical cyclone",
    "TCNTL": "Transcontinental",
    "TCU": "Towering cumulus",
    "TDA": "Today",
    "TEI": "Text element indicator",
    "TEMP": "Temperature",
    "TEMPO": "Temporary",
    "THD": "Thunderhead",
    "THDR": "Thunder",
    "THK": "Thick",
    "THKNG": "Thickening",
    "THKNS": "Thickness",
    "THKR": "Thicker",
    "THKST": "Thickest",
    "THN": "Thin",
    "THNG": "Thinning",
    "THNR": "Thinner",
    "THNST": "Thinnest",
    "THR": "Threshold",
    "THRFTR": "Thereafter",
    "THRU": "Through",
    "THRUT": "Throughout",
    "THSD": "Thousand",
    "THTN": "Threaten",
    "THTND": "Threatened",
    "THTNG": "Threatening",
    "THTNS": "Threatens",
    "TIL": "Until",
    "TKOF": "Takeoff",
    "TM": "Time",
    "TMPRY": "Temporary",
    "TMPRYLY": "Temporarily",
    "TMW": "Tomorrow",
    "TNDCY": "Tendency",
    "TNDCYS": "Tendencies",
    "TNGT": "Tonight",
    "TNTV": "Tentative",
    "TNTVLY": "Tentatively",
    "TOC": "Top of Climb",
    "TOP": "Top of Clouds",
    "TOPS": "Tops",
    "TOVC": "Top of overcast",
    "TP": "Type of aircraft",
    "TPG": "Topping",
    "TRBL": "Trouble",
    "TRIB": "Tributary",
    "TRKG": "Tracking",
    "TRML": "Terminal",
    "TRMT": "Terminate",
    "TRMTD": "Terminated",
    "TRMTG": "Terminating",
    "TRMTS": "Terminates",
    "TRNSP": "Transport",
    "TRNSPG": "Transporting",
    "TROF": "Trough",
    "TROFS": "Troughs",
    "TROP": "Tropopause",
    "TRPCD": "Tropical continental air mass",
    "TRPCL": "Tropical",
    "TRRN": "Terrain",
    "TRSN": "Transition",
    "TRW": "Thunderstorm",
    "TRW+": "Thunderstorm with heavy rain shower",
    "TS": "Thunderstorm",
    "TS+": "Thunderstorm with heavy snow",
    "TSFR": "Transfer",
    "TSFRD": "Transferred",
    "TSFRG": "Transferring",
    "TSFRS": "Transfers",
    "TSGR": "Thunderstorm with hail",
    "TSHWR": "Thundershower",
    "TSNO": "Thunderstorm information not available",
    "TSNT": "Transient",
    "TSQLS": "Thundersquall",
    "TSTM": "Thunderstorm",
    "TSW": "Thunderstorm with snow showers",
    "TSW+": "Thunderstorm with heavy snow showers",
    "TURB": "Turbulence",
    "TURBC": "Turbulence",
    "TURBT": "Turbulent",
    "TWD": "Toward",
    "TWDS": "Towards",
    "TWI": "Twilight",
    "TWR": "Tower",
    "TWRG": "Towering",
    "UA": "Pilot weather reports",
    "UDDF": "Up and downdrafts",
    "UIR": "Upper flight information region",
    "UN": "Unable",
    "UNAVBL": "Unavailable",
    "UNEC": "Unnecessary",
    "UNKN": "Unknown",
    "UNL": "Unlimited",
    "UNRELBL": "Unreliable",
    "UNRSTD": "Unrestricted",
    "UNSATFY": "Unsatisfactory",
    "UNSBL": "Unseasonable",
    "UNSTBL": "Unstable",
    "UNSTDY": "Unsteady",
    "UNSTL": "Unsettle",
    "UNSTLD": "Unsettled",
    "UNUSBL": "Unusable",
    "UP": "Unknown precipitation",
    "UPDFTS": "Updrafts",
    "UPR": "Upper",
    "UPSLP": "Upslope",
    "UPSTRM": "Upstream",
    "URG": "Urgent",
    "USBL": "Usable",
    "UTC": "Universal Time Coordinate",
    "UUA": "Urgent PIREP Weather Reports",
    "UVV": "Upward vertical velocity",
    "UWNDS": "Upper winds Varies",
    "V": "Varies",
    "VA": "Volcanic Ash",
    "VAAC": "Volcanic Ash Advisory Center",
    "VAAS": "Volcanic Ash Advisory Statement",
    "VAD": "Velocity azimuth display",
    "VAL": "Valley",
    "VARN": "Variation",
    "VC": "Vicinity",
    "VCNTY": "Vicinity",
    "VCOT": "VFR conditions on top",
    "VCTR": "Vector",
    "VCTS": "Thunderstorms in the vicinity",
    "VDUC": "VAS Data Utilization Center",
    "VFR": "Visual flight rules",
    "VFY": "Verify",
    "VFYD": "Verified",
    "VFYG": "Verifying",
    "VFYS": "Verifies",
    "VHF": "Very High Frequency",
    "VIS": "Visibility",
    "VLCTY": "Velocity",
    "VLCTYS": "Velocities",
    "VLNT": "Violent",
    "VLNTLY": "Violently",
    "VLY": "Valley",
    "VMC": "Visual meteorological conditions",
    "VOL": "Volume",
    "VOR": "VHF Omnidirectional Radio Range",
    "VORT": "Vorticity",
    "VORTAC": "VOR and TACAN combination",
    "VR": "Veer",
    "VRB": "Variable",
    "VRBL": "Variable",
    "VRG": "Veering",
    "VRS": "Veers",
    "VRT MOTN": "Vertical motion",
    "VRY": "Very",
    "VSB": "Visible",
    "VSBY": "Visibility",
    "VSBYDR": "Visibility decreasing rapidly",
    "VSBYIR": "Visibility increasing rapidly",
    "VSNO": "Visibility at secondary location not available",
    "VV": "Vertical visibility",
    "VWP": "VAD Wind profiler",
    "W": "West",
    "WAA": "Warm air advection",
    "WAFS": "Word Area Forecast System",
    "WBND": "Westbound",
    "WDLY": "Widely",
    "WDSPRD": "Widespread",
    "WEA": "Weather",
    "WFO": "Weather Forecast Office",
    "WFP": "Warm front passage",
    "WFSO": "Weather Forecast Service Office",
    "WI": "Within",
    "WIBIS": "Will be issued",
    "WID": "Width",
    "WINT": "Winter",
    "WK": "Weak",
    "WKDAY": "Weekday",
    "WKEND": "Weekend",
    "WKN": "Weaken",
    "WKNG": "Weakening",
    "WKNS": "Weakens",
    "WKR": "Weaker",
    "WKST": "Weakest",
    "WL": "Will",
    "WLY": "Westerly",
    "WMO": "World Meteorological Organization",
    "WND": "Wind",
    "WNDS": "Winds",
    "WNW": "West-northwest",
    "WNWLY": "West-northwesterly",
    "WNWRN": "West-northwestern",
    "WNWWD": "West-northwestward",
    "WO": "Without",
    "WRM": "Warm",
    "WRMFNT": "Warm front",
    "WRMFNTL": "Warm frontal",
    "WRMG": "Warming",
    "WRMR": "Warmer",
    "WRMST": "Warmest",
    "WRN": "Western",
    "WRNG": "Warning",
    "WRS": "Worse",
    "WS": "Wind shear",
    "WSFO": "Weather Service Forecast Office",
    "WSHFT": "Windshift",
    "WSO": "Weather service office",
    "WSPD": "Wind speed",
    "WSR-88D": "NWS Doppler Radar",
    "WSW": "West-southwest",
    "WSWLY": "West-southwesterly",
    "WSWRN": "West-southwestern",
    "WSWWD": "West-southwestward",
    "WTR": "Water",
    "WTSPT": "Waterspout",
    "WUD": "Would",
    "WV": "Wind",
    "WVS": "Waves",
    "WW": "Severe weather watch",
    "WWAMKC": "SPC status report",
    "WWD": "Westward",
    "WWS": "Severe weather watches",
    "WX": "Weather",
    "XCP": "Except",
    "XPC": "Expect",
    "XPCD": "Expected",
    "XPCG": "Expecting",
    "XPCS": "Expects",
    "XPLOS": "Explosive",
    "XTND": "Extend",
    "XTNDD": "Extended",
    "XTNDG": "Extending",
    "XTRM": "Extreme",
    "XTRMLY": "Extremely",
    "YDA": "Yesterday",
    "Z": "Zulu time",
    "ZL": "Freezing drizzle",
    "ZN": "Zone",
    "ZNS": "Zones",
    "ZR": "Freezing rain",
}



"""
Methods to resolve flight paths in coordinates
"""

# stdlib
from typing import List, Optional, Union

# library
from geopy.distance import great_circle  # type: ignore

# module
from avwx.exceptions import BadStation
from avwx.load_utils import LazyLoad
from avwx.station import Station
from avwx.structs import Coord

NAVAIDS = LazyLoad("navaids")
QCoord = Union[Coord, List[Coord]]


def _distance(near: Coord, far: Coord) -> float:
    return great_circle(near.pair, far.pair).nm


def _closest(coord: QCoord, coords: List[Coord]) -> Coord:
    if isinstance(coord, Coord):
        distances = [(_distance(coord, c), c) for c in coords]
    else:
        distances = [(_distance(c, _closest(c, coords)), c) for c in coord]
    distances.sort(key=lambda x: x[0])
    return distances[0][1]


def _best_coord(
    previous: Optional[QCoord],
    current: QCoord,
    up_next: Optional[QCoord],
) -> Coord:
    """Determine the best coordinate based on surroundings
    At least one of these should be a list
    """
    if previous is None and up_next is None:
        if isinstance(current, list):
            raise Exception("Unable to determine best coordinate")
        return current
    # NOTE: add handling to determine best midpoint
    if up_next is None:
        up_next = previous
    if isinstance(up_next, list):
        return _closest(current, up_next)
    return _closest(up_next, current)  # type: ignore


def to_coordinates(
    values: List[Union[Coord, str]], last_value: Optional[QCoord] = None
) -> List[Coord]:
    """Convert any known idents found in a flight path into coordinates

    Prefers Coord > ICAO > Navaid > IATA
    """
    if not values:
        return []
    coord = values[0]
    if isinstance(coord, str):
        try:
            value = coord
            coord = Station.from_icao(coord).coord
            coord.repr = value
        except BadStation:
            try:
                coords = [Coord(lat=c[0], lon=c[1], repr=coord) for c in NAVAIDS[coord]]  # type: ignore
            except KeyError:
                value = coord
                coord = Station.from_iata(coord).coord  # type: ignore
                coord.repr = value
            else:
                if len(coords) == 1:
                    coord = coords[0]
                else:
                    new_coords = to_coordinates(values[1:], coords)
                    new_coord = new_coords[0] if new_coords else None
                    coord = _best_coord(last_value, coords, new_coord)
                    return [coord] + new_coords
    return [coord] + to_coordinates(values[1:], coord)



"""
Builds the main station list

Source file for airports.csv and runways.csv can be downloaded from
http://ourairports.com/data/

Source file for stations.txt can be downloaded from
https://www.aviationweather.gov/docs/metar/stations.txt
"""

# stdlib
import csv
import json
import logging
from contextlib import suppress
from datetime import date
from pathlib import Path
from typing import Dict, List, Optional

# library
import httpx

# module
from avwx.data.mappers import FILE_REPLACE, SURFACE_TYPES


LOG = logging.getLogger("avwx.data.build_stations")


def load_stations(path: Path) -> set[str]:
    """Load a station set from a path"""
    return set(path.read_text().strip().split("\n"))


_FILE_DIR = Path(__file__).parent
_DATA = _FILE_DIR / "files"
GOOD_PATH = _DATA / "good_stations.txt"
OUTPUT_PATH = _DATA / "stations.json"


DATA_ROOT = "https://davidmegginson.github.io/ourairports-data/"
_SOURCE: Dict[str, str] = {}
_SOURCES = {
    "airports": DATA_ROOT + "airports.csv",
    "runways": DATA_ROOT + "runways.csv",
    "stations": "https://www.aviationweather.gov/docs/metar/stations.txt",
}


ACCEPTED_STATION_TYPES = [
    "balloonport",
    # "closed",
    "heliport",
    "large_airport",
    "medium_airport",
    "seaplane_base",
    "small_airport",
]


def nullify(data: dict) -> dict:
    """Nullify empty strings in a dict"""
    for key, val in data.items():
        if isinstance(val, str) and not val.strip():
            data[key] = None
    return data


def format_coord(coord: str) -> float:
    """Convert coord string to float"""
    neg = -1 if coord[-1] in ("S", "W") else 1
    return neg * float(coord[:-1].strip().replace(" ", "."))


def validate_icao(code: str) -> Optional[str]:
    """Validates a given station ident"""
    if len(code) != 4:
        return None
    return code.upper()


def get_icao(station: List[str]) -> Optional[str]:
    """Finds the ICAO by checking ident and GPS code"""
    return validate_icao(station[12]) or validate_icao(station[1])


def clean_source_files():
    """Cleans the source data files before parsing"""
    text = _SOURCE["airports"]
    for find, replace in FILE_REPLACE.items():
        text = text.replace(find, replace)
    _SOURCE["airports"] = text


def format_station(icao: str, station: List[str]) -> dict:
    """Converts source station list into info dict"""
    try:
        elev_ft = float(station[6])
        elev_m = round(elev_ft * 0.3048)
        elev_ft = round(elev_ft)
    except ValueError:
        elev_ft, elev_m = None, None
    index = station[9].find("-")
    ret = {
        "type": station[2],
        "name": station[3],
        "reporting": None,
        "latitude": float(station[4]),
        "longitude": float(station[5]),
        "elevation_ft": elev_ft,
        "elevation_m": elev_m,
        "country": station[9][:index],
        "state": station[9][index + 1 :],
        "city": station[10],
        "icao": icao,
        "iata": station[13].upper(),
        "website": station[15],
        "wiki": station[16],
        "note": station[17],
    }
    return nullify(ret)


def build_stations() -> tuple[dict, dict]:
    """Builds the station dict from source file"""
    stations, icao_map = {}, {}
    data = csv.reader(_SOURCE["airports"].splitlines())
    next(data)  # Skip header
    for station in data:
        icao = get_icao(station)
        if icao and station[2] in ACCEPTED_STATION_TYPES:
            stations[icao] = format_station(icao, station)
            icao_map[station[0]] = icao
    return stations, icao_map


def add_missing_stations(stations: dict) -> dict:
    """Add non-airport stations from NOAA"""
    for line in _SOURCE["stations"].splitlines():
        # Must be data line with METAR reporting
        if len(line) != 84 or line[0] == "!" or line[62] != "X":
            continue
        icao = line[20:24].strip().upper()
        if not icao or icao in stations:  # or icao in BAD_STATIONS:
            continue
        elev_m = int(line[55:59].strip())
        ret = {
            "type": "weather_station",
            "name": line[3:19].strip(),
            "reporting": None,
            "latitude": format_coord(line[39:45]),
            "longitude": format_coord(line[47:54]),
            "elevation_ft": round(elev_m * 3.28084),
            "elevation_m": elev_m,
            "country": line[81:83].strip(),
            "state": line[:2],
            "city": None,
            "icao": icao,
            "iata": line[26:29].strip().upper(),
            "website": None,
            "wiki": None,
            "note": None,
        }
        stations[icao] = nullify(ret)
    return stations


def get_surface_type(surface: str) -> Optional[str]:
    """Returns the normalize surface type value"""
    for key, items in SURFACE_TYPES.items():
        if surface in items:
            return key
    return None


def add_runways(stations: dict, icao_map: dict) -> dict:
    """Add runway information to station if availabale"""
    data = csv.reader(_SOURCE["runways"].splitlines())
    next(data)  # Skip header
    for runway in data:
        # if runway is closed
        if runway[7] != "0":
            continue
        out = {
            "length_ft": int(runway[3]) if runway[3] else 0,
            "width_ft": int(runway[4]) if runway[4] else 0,
            "surface": get_surface_type(runway[5].lower()),
            "lights": runway[6] == "1",
            "ident1": runway[8],
            "ident2": runway[14],
            "bearing1": float(runway[12]) if runway[12] else None,
            "bearing2": float(runway[18]) if runway[18] else None,
        }
        icao = icao_map.get(runway[1], runway[2])
        with suppress(KeyError):
            if "runways" in stations[icao]:
                stations[icao]["runways"].append(out)
            else:
                stations[icao]["runways"] = [out]
    # Sort runways by longest length and add missing nulls
    for icao in stations:
        if "runways" in stations[icao]:
            stations[icao]["runways"].sort(key=lambda x: x["length_ft"], reverse=True)
        else:
            stations[icao]["runways"] = None
    return stations


def add_reporting(stations: dict) -> dict:
    """Add reporting boolean to station if available"""
    good = load_stations(GOOD_PATH)
    for icao in stations:
        stations[icao]["reporting"] = icao in good
    return stations


def download_source_files() -> bool:
    """Returns True if source files updated successfully"""
    for key, route in _SOURCES.items():
        resp = httpx.get(route)
        if resp.status_code != 200:
            return False
        _SOURCE[key] = resp.text
    return True


def update_station_info_date():
    """Update the package's station meta date"""
    meta_path = _FILE_DIR.parent / "station" / "meta.py"
    meta = meta_path.open().read()
    target = '__LAST_UPDATED__ = "'
    start = meta.find(target) + len(target)
    prefix = meta[:start]
    end = start + 10
    output = prefix + date.today().strftime(r"%Y-%m-%d") + meta[end:]
    with meta_path.open("w") as out:
        out.write(output)


def main() -> int:
    """Build/update the stations.json main file"""
    LOG.info("Fetching")
    if not download_source_files():
        LOG.error("Unable to update source files")
        return 1
    LOG.info("Cleaning")
    clean_source_files()
    LOG.info("Building")
    stations, icao_map = build_stations()
    stations = add_missing_stations(stations)
    stations = add_reporting(stations)
    stations = add_runways(stations, icao_map)
    LOG.info("Saving")
    json.dump(
        stations,
        OUTPUT_PATH.open("w", encoding="utf8"),
        sort_keys=True,
        indent=1,
        ensure_ascii=False,
    )
    LOG.info("Updating station date")
    update_station_info_date()
    return 0


if __name__ == "__main__":
    LOG.setLevel("info")
    main()



"""
AIRMET / SIGMET static vales
"""

AIRMET_KEY = "airmet"

BULLETIN_TYPES = {
    "WS": "sigmet",
    "WV": "volcanic ash",
    "WC": "tropical cyclone",
    "WA": AIRMET_KEY,
}

INTENSITY = {
    "NC": "No change",
    "INTSF": "Intensifying",
    "WKN": "Weakening",
}

# More to less specific when overlapping
WEATHER_TYPES = {
    "DMSHG LINE TS": "Diminishing line of thunderstorms",
    "AREA TS": "Thunderstorms in the area",
    "LINE TS": "Line of thunderstorms",
    "OBSC TS": "Obscured thunderstorms",
    "EMBD TS": "Embedded thunderstorms",
    "FRQ TS": "Frequent thunderstorms",
    "SQL TS": "Squall line thunderstorms",
    "OBSC TSGR": "Obscured thunderstorms with hail",
    "EMBD TSGR": "Embedded thunderstorms with hail",
    "FRQ TSGR": "Frequent thunderstorms with hail",
    "SQL TSGR": "Squall line thunderstorms with hail",
    "MOD TURB": "Moderate Turbulence",
    "SEV TURB": "Severe turbulence",
    "TURB": "Turbulence",
    "MOD ICE (FZRA)": "Moderate icing due to freezing rain",
    "SEV ICE (FZRA)": "Severe icing due to freezing rain",
    "MOD ICE": "Moderate icing",
    "SEV ICE": "Severe icing",
    "SEV MTW": "Severe mountain wave",
    "HVY DS": "Heavy duststorm",
    "HVY SS": "Heavy sandstorm",
    "RDOACT CLD": "Radioactive cloud",
    "VA ERUPTION": "Volcanic eruption",
    "VA CLD": "Volcanic cloud",
    "MTNS OBSC": "Mountain Obscuration",
    "MTNS OBSCN": "Mountain Obscuration",
    "CB": "Cumulonimbus",
}



"""
AIRMET / SIGMET report parsing
"""

# stdlib
import asyncio as aio
import re
from contextlib import suppress
from datetime import date, datetime, timezone
from itertools import chain
from typing import List, Optional, Tuple

# library
from geopy.distance import distance as geo_distance  # type: ignore

# module
from avwx.base import AVWXBase
from avwx.flight_path import to_coordinates
from avwx.load_utils import LazyLoad
from avwx.parsing import core
from avwx.service.bulk import NOAA_Bulk, NOAA_Intl, Service
from avwx.static.core import CARDINAL_DEGREES, IN_UNITS
from avwx.static.airsigmet import BULLETIN_TYPES, INTENSITY, WEATHER_TYPES
from avwx.structs import (
    AirSigmetData,
    Bulletin,
    Code,
    Coord,
    Movement,
    Number,
    AirSigObservation,
    Timestamp,
    Units,
)

# N1429 W09053 - N1427 W09052 - N1411 W09139 - N1417 W09141
_COORD_PATTERN = re.compile(r"\b[NS]\d{4} [EW]\d{5}\b( -)?")

# FROM 60NW ISN-INL-TVC-GIJ-UIN-FSD-BIL-60NW ISN
# FROM 70SSW ISN TO 20NNW FAR TO 70E DLH TO 40SE EAU TO 80SE RAP TO 40NNW BFF TO 70SSW
_NAVAID_PATTERN = re.compile(
    r"\b(\d{1,3}[NESW]{1,3} [A-z]{3}-?\b)|((-|(TO )|(FROM ))[A-z]{3}\b)"
)

# N OF N2050 AND S OF N2900
_LATTERAL_PATTERN = re.compile(
    r"\b([NS] OF [NS]\d{2,4})|([EW] OF [EW]\d{3,5})( AND)?\b"
)

NAVAIDS = LazyLoad("navaids")

# Used to assist parsing after sanitized. Removed after parse
_FLAGS = {
    "...": " <elip> ",
    "..": " <elip> ",
    ". ": " <break> ",
    "/VIS ": " <vis> VIS ",
}


def _parse_prep(report: str) -> List[str]:
    """Prepares sanitized string by replacing elements with flags"""
    report = report.rstrip(".")
    for key, val in _FLAGS.items():
        report = report.replace(key, val)
    return report.split()


def _clean_flags(data: List[str]) -> List[str]:
    return [i for i in data if i[0] != "<"]


def _bulletin(value: str) -> Bulletin:
    # if len(value) != 6:
    #     return None
    type_key = value[:2]
    return Bulletin(
        repr=value,
        type=Code(repr=type_key, value=BULLETIN_TYPES[type_key]),
        country=value[2:4],
        number=int(value[4:]),
    )


def _header(data: List[str]) -> Tuple[List[str], Bulletin, str, str, Optional[str]]:
    bulletin = _bulletin(data[0])
    correction, end = None, 3
    if len(data[3]) == 3:
        correction, end = data[3], 4
    return data[end:], bulletin, data[1], data[2], correction


def _spacetime(
    data: List[str],
) -> Tuple[List[str], str, str, Optional[str], str, Optional[str]]:
    area = data.pop(0)
    # Skip airmet type + time repeat
    if data[0] == "WA" and data[1].isdigit():
        data = data[2:]
        area = area[:-1]  # Remove type label from 3-letter ident
    valid_index = data.index("VALID")
    report_type = " ".join(data[:valid_index])
    data = data[valid_index + 1 :]
    if data[0] == "UNTIL":
        start_time = None
        end_time = data[1]
        data = data[2:]
    else:
        target = "-" if "-" in data[0] else "/"
        start_time, end_time = data.pop(0).split(target)
    # KMCO- ORL FIR
    if data[0][-1] == "-":
        station = data.pop(0)[:-1]
    # KMCO - KMCO
    elif data[1] == "-" and len(data[0]) == 4:
        station = data.pop(0)
        data.pop(0)
    else:
        station = None
    return data, area, report_type, start_time, end_time, station


def _first_index(data: List[str], *targets: str) -> int:
    for target in targets:
        with suppress(ValueError):
            return data.index(target)
    return -1


def _region(data: List[str]) -> Tuple[List[str], str]:
    # FIR/CTA region name
    name_end = _first_index(data, "FIR", "CTA") + 1
    # Non-standard name using lookahead Ex: FL CSTL WTRS FROM 100SSW
    if not name_end:
        name_end = _first_index(data, "FROM")
    # State list
    if not name_end:
        for item in data:
            if len(item) == 2:
                name_end += 1
            else:
                break
    name = " ".join(data[:name_end])
    return data[name_end:], name


def _time(
    data: List[str], issued: date = None
) -> Tuple[List[str], Optional[Timestamp], Optional[Timestamp]]:
    """Extracts the start and/or end time based on a couple starting elements"""
    index = _first_index(data, "AT", "FCST", "UNTIL", "VALID", "OUTLOOK", "OTLK")
    if index == -1:
        return data, None, None
    start_item = data.pop(index)
    start, end, observed = None, None, None
    if "-" in data[index]:
        start_item, end_item = data.pop(index).split("-")
        start = core.make_timestamp(
            start_item, time_only=len(start_item) < 6, target_date=issued
        )
        end = core.make_timestamp(
            end_item, time_only=len(end_item) < 6, target_date=issued
        )
    elif len(data[index]) >= 4 and data[index][:4].isdigit():
        observed = core.make_timestamp(
            data.pop(index), time_only=True, target_date=issued
        )
        if index > 0 and data[index - 1] == "OBS":
            data.pop(index - 1)
    for remv in ("FCST", "OUTLOOK", "OTLK", "VALID"):
        with suppress(ValueError):
            data.remove(remv)
    if observed:
        if start_item in ("UNTIL", "VALID"):
            end = observed
        else:
            start = observed
    return data, start, end


def _coord_value(value: str) -> float:
    if value[0] in ("N", "S"):
        index, strip, replace = 3, "N", "S"
    else:
        index, strip, replace = 4, "E", "W"
    num = f"{value[:index]}.{value[index:]}".lstrip(strip).replace(replace, "-")
    return float(num)


def _position(data: List[str]) -> Tuple[List[str], Optional[Coord]]:
    try:
        index = data.index("PSN")
    except ValueError:
        return data, None
    data.pop(index)
    raw = f"{data[index]} {data[index + 1]}"
    lat = _coord_value(data.pop(index))
    lon = _coord_value(data.pop(index))
    return data, Coord(lat=lat, lon=lon, repr=raw)


def _movement(
    data: List[str], units: Units
) -> Tuple[List[str], Units, Optional[Movement]]:
    with suppress(ValueError):
        data.remove("STNR")
        speed = core.make_number("STNR")
        return data, units, Movement(repr="STNR", direction=None, speed=speed)
    try:
        index = data.index("MOV")
    except ValueError:
        return data, units, None
    raw = data.pop(index)
    direction_str = data.pop(index)
    # MOV CNL
    if direction_str == "CNL":
        return data, units, None
    raw += " " + direction_str + " "
    # MOV FROM 23040KT
    if direction_str == "FROM":
        value = data[index][:3]
        raw += value
        direction = core.make_number(value)
        data[index] = data[index][3:]
    # MOV E 45KMH
    else:
        direction = core.make_number(
            direction_str.replace("/", ""), literal=True, special=CARDINAL_DEGREES
        )
    speed = None
    with suppress(IndexError):
        kt_unit, kmh_unit = data[index].endswith("KT"), data[index].endswith("KMH")
        if kt_unit or kmh_unit:
            units.wind_speed = "kmh" if kmh_unit else "kt"
            speed_str = data.pop(index)
            raw += speed_str
            # Remove bottom speed Ex: MOV W 05-10KT
            if "-" in speed_str:
                speed_str = speed_str[speed_str.find("-") + 1 :]
            speed = core.make_number(speed_str[: -3 if kmh_unit else -2])
    return data, units, Movement(repr=raw.strip(), direction=direction, speed=speed)


def _info_from_match(match: re.Match, start: int) -> Tuple[str, int]:
    """Returns the matching text and starting location if none yet available"""
    if start == -1:
        start = match.start()
    return match.group(), start


def _pre_break(report: str) -> str:
    break_index = report.find(" <break> ")
    if break_index != -1:
        return report[:break_index]
    return report


def _bounds_from_latterals(report: str, start: int) -> Tuple[str, List[str], int]:
    """Extract coordinate latterals from report Ex: N OF N2050"""
    bounds = []
    for match in _LATTERAL_PATTERN.finditer(_pre_break(report)):
        group, start = _info_from_match(match, start)
        # post 3.8 bounds.append(group.removesuffix(" AND"))
        if group.endswith(" AND"):
            group = group[:-4]
        bounds.append(group)
        report = report.replace(group, " ")
    return report, bounds, start


def _coords_from_text(report: str, start: int) -> Tuple[str, List[Coord], int]:
    """Extract raw coordinate values from report Ex: N4409 E01506"""
    coords = []
    for match in _COORD_PATTERN.finditer(_pre_break(report)):
        group, start = _info_from_match(match, start)
        text = group.strip(" -")
        lat, lon = text.split()
        coord = Coord(lat=_coord_value(lat), lon=_coord_value(lon), repr=text)
        coords.append(coord)
        report = report.replace(group, " ")
    return report, coords, start


def _coords_from_navaids(report: str, start: int) -> Tuple[str, List[Coord], int]:
    """Extract navaid referenced coordinates from report Ex: 30SSW BNA"""
    # pylint: disable=too-many-locals
    coords, navs = [], []
    for match in _NAVAID_PATTERN.finditer(_pre_break(report)):
        group, start = _info_from_match(match, start)
        report = report.replace(group, " ")
        group = group.strip("-")  # post 3.8 .removeprefix("FROM ").removeprefix("TO ")
        for end in ("FROM", "TO"):
            if group.startswith(end + " "):
                group = group[(len(end) + 1) :]
        navs.append((group, *group.split()))
    locs = to_coordinates([n[2 if len(n) == 3 else 1] for n in navs])
    for i, nav in enumerate(navs):
        value = nav[0]
        if len(nav) == 3:
            vector, num_index = nav[1], 0
            while vector[num_index].isdigit():
                num_index += 1
            distance, bearing = (
                int(vector[:num_index]),
                CARDINAL_DEGREES[vector[num_index:]],
            )
            loc = geo_distance(nautical=distance).destination(
                locs[i].pair, bearing=bearing
            )
            coord = Coord(lat=loc.latitude, lon=loc.longitude, repr=value)
        else:
            coord = locs[i]
            coord.repr = value
        coords.append(coord)
    return report, coords, start


def _bounds(data: List[str]) -> Tuple[List[str], List[Coord], List[str]]:
    """Extract coordinate bounds by coord, navaid, and latterals"""
    report, start = " ".join(data), -1
    report, bounds, start = _bounds_from_latterals(report, start)
    report, coords, start = _coords_from_text(report, start)
    report, navs, start = _coords_from_navaids(report, start)
    coords += navs
    for target in ("FROM", "WI", "BOUNDED", "OBS"):
        index = report.find(target + " ")
        if index != -1 and index < start:
            start = index
    report = report[:start] + report[report.rfind("  ") :]
    data = [s for s in report.split() if s]
    return data, coords, bounds


def _is_altitude(value: str) -> bool:
    """Returns True if the value is a possible altitude"""
    if len(value) < 5:
        return False
    if value[:4] == "SFC/":
        return True
    if value[:2] == "FL" and value[2:5].isdigit():
        return True
    first, *_ = value.split("/")
    if first[-2:] == "FT" and first[-5:-2].isdigit():
        return True
    return False


def _make_altitude(
    value: str, units: Units, force_fl: bool = False
) -> Tuple[Optional[Number], Units]:
    """Convert altitude string into a number"""
    raw = value
    for end in ("FT", "M"):
        if value.endswith(end):
            force_fl = False
            units.altitude = end.lower()
            # post 3.8 value = value.removesuffix(end)
            value = value[: -len(end)]
    if force_fl:
        value = "FL" + value
    return core.make_number(value, repr=raw), units


def _altitudes(
    data: List[str], units: Units
) -> Tuple[List[str], Units, Optional[Number], Optional[Number]]:
    """Extract the floor and ceiling altitudes"""
    floor, ceiling = None, None
    for i, item in enumerate(data):
        # BTN FL180 AND FL330
        if item == "BTN" and len(data) > i + 2 and data[i + 2] == "AND":
            floor, units = _make_altitude(data[i + 1], units)
            ceiling, units = _make_altitude(data[i + 3], units)
            data = data[:i] + data[i + 4 :]
            break
        # TOPS ABV FL450
        if item in ("TOP", "TOPS", "BLW"):
            if data[i + 1] == "ABV":
                ceiling = core.make_number("ABV " + data[i + 2])
                data = data[:i] + data[i + 3 :]
                break
            # TOPS TO FL310
            if data[i + 1] == "TO":
                data.pop(i)
            ceiling, units = _make_altitude(data[i + 1], units)
            data = data[:i] + data[i + 2 :]
            # CIG BLW 010
            if data[i - 1] == "CIG":
                data.pop(i - 1)
            break
        # FL060/300 SFC/FL160
        if _is_altitude(item):
            if "/" in item:
                floor_val, ceiling_val = item.split("/")
                floor, units = _make_altitude(floor_val, units)
                if (floor_val == "SFC" or floor_val[:2] == "FL") and ceiling_val[
                    :2
                ] != "FL":
                    ceiling, units = _make_altitude(ceiling_val, units, True)
                else:
                    ceiling, units = _make_altitude(ceiling_val, units)
            else:
                ceiling, units = _make_altitude(item, units)
            data.pop(i)
            break
    return data, units, floor, ceiling


def _weather_type(data: List[str]) -> Tuple[List[str], Optional[Code]]:
    weather = None
    report = " ".join(data)
    for key, val in WEATHER_TYPES.items():
        if key in report:
            weather = Code(repr=key, value=val)
            data = [i for i in report.replace(key, "").split() if i]
            break
    return data, weather


def _intensity(data: List[str]) -> Tuple[List[str], Optional[Code]]:
    if not data:
        return data, None
    try:
        value = INTENSITY[data[-1]]
        code = data.pop()
        return data, Code(repr=code, value=value)
    except KeyError:
        return data, None


def _sigmet_observation(
    data: List[str], units: Units, issued: date = None
) -> Tuple[AirSigObservation, Units]:
    data, start_time, end_time = _time(data, issued)
    data, position = _position(data)
    data, coords, bounds = _bounds(data)
    data, units, movement = _movement(data, units)
    data, intensity = _intensity(data)
    data, units, floor, ceiling = _altitudes(data, units)
    data, weather = _weather_type(data)
    struct = AirSigObservation(
        type=weather,
        start_time=start_time,
        end_time=end_time,
        position=position,
        floor=floor,
        ceiling=ceiling,
        coords=coords,
        bounds=bounds,
        movement=movement,
        intensity=intensity,
        other=_clean_flags(data),
    )
    return struct, units


def _observations(
    data: List[str], units: Units, issued: date = None
) -> Tuple[Units, Optional[AirSigObservation], Optional[AirSigObservation]]:
    observation, forecast, forecast_index = None, None, -1
    forecast_index = _first_index(data, "FCST", "OUTLOOK", "OTLK")
    if forecast_index == -1:
        observation, units = _sigmet_observation(data, units, issued)
    # 6 is arbitrary. Will likely change or be more precise later
    elif forecast_index < 6:
        forecast, units = _sigmet_observation(data, units, issued)
    else:
        observation, units = _sigmet_observation(data[:forecast_index], units, issued)
        forecast, units = _sigmet_observation(data[forecast_index:], units, issued)
    return units, observation, forecast


_REPLACE = {" MO V ": " MOV "}


def sanitize(report: str) -> str:
    """Sanitized AIRMET / SIGMET report string"""
    report = report.strip(" =")
    for key, val in _REPLACE.items():
        report = report.replace(key, val)
    data = report.split()
    for i, item in reversed(list(enumerate(data))):
        # Remove extra element on altitude Ex: FL450Z skip 1000FT
        if (
            len(item) > 4
            and not item[-1].isdigit()
            and item[-2:] != "FT"
            and item[-1] != "M"
            and _is_altitude(item[:-1])
        ):
            data[i] = item[:-1]
    return " ".join(data)


def parse(report: str, issued: date = None) -> Tuple[AirSigmetData, Units]:
    """Parse AIRMET / SIGMET report string"""
    # pylint: disable=too-many-locals
    units = Units(**IN_UNITS)
    sanitized = sanitize(report)
    data, bulletin, issuer, time, correction = _header(_parse_prep(sanitized))
    data, area, report_type, start_time, end_time, station = _spacetime(data)
    body = sanitized[sanitized.find(" ".join(data[:2])) :]
    # Trim AIRMET type
    if data[0] == "AIRMET":
        data = data[data.index("<elip>") :]
    data, region = _region(data)
    units, observation, forecast = _observations(data, units, issued)
    struct = AirSigmetData(
        raw=report,
        sanitized=sanitized,
        station=station,
        time=core.make_timestamp(time, target_date=issued),
        remarks=None,
        bulletin=bulletin,
        issuer=issuer,
        correction=correction,
        area=area,
        type=report_type,
        start_time=core.make_timestamp(start_time, target_date=issued),
        end_time=core.make_timestamp(end_time, target_date=issued),
        body=body,
        region=region,
        observation=observation,
        forecast=forecast,
    )
    return struct, units


class AirSigmet(AVWXBase):
    """Class representing an AIRMET or SIGMET report"""

    data: Optional[AirSigmetData] = None

    def _post_parse(self) -> None:
        if self.raw:
            self.data, self.units = parse(self.raw, self.issued)

    @staticmethod
    def sanitize(report: str) -> str:
        """Sanitizes the report string"""
        return sanitize(report)


class AirSigManager:
    """Class to fetch and manage AIRMET and SIGMET reports"""

    _services: List[Service]  # type: ignore
    _raw: List[Tuple[str, str]]
    last_updated: Optional[datetime] = None
    raw: List[str]
    reports: Optional[List[AirSigmet]] = None

    def __init__(self):
        self._services = [NOAA_Bulk("airsigmet"), NOAA_Intl("airsigmet")]
        self._raw, self.raw = [], []

    async def _update(
        self, index: int, timeout: int
    ) -> List[Tuple[str, Optional[str]]]:
        source = self._services[index].root
        reports = await self._services[index].async_fetch(timeout=timeout)  # type: ignore
        raw: List[Tuple[str, Optional[str]]] = []
        for report in reports:
            if not report:
                continue
            raw.append((report, source))
        return raw

    def update(self, timeout: int = 10, disable_post: bool = False) -> bool:
        """Updates fetched reports and returns whether they've changed"""
        return aio.run(self.async_update(timeout, disable_post))

    async def async_update(self, timeout: int = 10, disable_post: bool = False) -> bool:
        """Updates fetched reports and returns whether they've changed"""
        coros = [self._update(i, timeout) for i in range(len(self._services))]
        data = await aio.gather(*coros)
        raw = list(chain.from_iterable(data))
        reports = [i[0] for i in raw]
        changed = raw != self.raw
        if changed:
            self._raw, self.raw = raw, reports
            self.last_updated = datetime.now(tz=timezone.utc)
        # Parse reports if not disabled
        if not disable_post:
            parsed = []
            for report, source in raw:
                if obj := AirSigmet.from_report(report):
                    obj.source = source
                    parsed.append(obj)
            self.reports = parsed
        return changed



"""
General aviation data fetch and updates
"""

from avwx.data.build_aircraft import main as update_aircraft
from avwx.data.build_navaids import main as update_navaids
from avwx.data.build_stations import main as update_stations


def update_all() -> bool:
    """Update all local data. Requires a reimport to guarentee update"""
    for func in (update_aircraft, update_navaids, update_stations):
        if func():
            return False
    return True



"""
avwx Package Setup
"""

from setuptools import find_namespace_packages, setup

VERSION = "1.7.1"

dependencies = [
    "geopy~=2.2",
    "httpx~=0.22",
    "python-dateutil~=2.8",
    "xmltodict~=0.12",
]

test_dependencies = ["pytest-asyncio~=0.18", "time-machine~=2.6"]

extras = {
    "fuzz": ["rapidfuzz~=2.0"],
    "scipy": ["scipy~=1.8"],
    "docs": ["mkdocs~=1.3", "mkdocs-material~=8.2", "mkdocs-minify-plugin~=0.5"],
    "tests": test_dependencies,
}
extras["all"] = extras["fuzz"] + extras["scipy"]

setup(
    name="avwx-engine",
    version=VERSION,
    description="Aviation weather report parsing library",
    url="https://github.com/avwx-rest/avwx-engine",
    author="Michael duPont",
    author_email="michael@dupont.dev",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">= 3.8",
    install_requires=dependencies,
    packages=find_namespace_packages(include=["avwx*"]),
    package_data={
        "avwx.data.files": [
            "aircraft.json",
            "good_stations.txt",
            "navaids.json",
            "stations.json",
        ]
    },
    tests_require=test_dependencies,
    extras_require=extras,
)
